/*
 * hashsyn.c - count tcp syns with mention of rst/fin on a per ip src basis. 
 */

/*
 * Copyright (c) 2004 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse
 *    or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <syslog.h>

#include <pcap.h>

/* use filter.h which is more general because we access
 * topn_syn in here
 */

/* 
#include "hashsyn.h"
*/
#include "ourmon.h"
#include "filter.h"
#include "trigger.h"
#include "ehash.c"	/* inlined function */

/* format string for printing out syn tuple
 * N items PLUS port tuples. includes port count.
*/
#define SYNFORMAT "%s:%u:%u:%u:%u:%u:%u:%u:%u:%x:%u:%u:%u:%u:" 

static 
void __inline
printSynTuple(FILE *fd, struct synnode *head, int i, int scanflag, struct slist *isl, struct slist *tsl)
{
	int ipds = 0; 	
	int tcpds = 0;
	if (scanflag) {
		ipds = scanLookup(isl, head[i].ipsrc);
		tcpds = scanLookup(tsl, head[i].ipsrc);
	}
	fprintf(fd, SYNFORMAT,
		intoa(head[i].ipsrc), 			/* 0 */
		head[i].syncount, 			/* 1 */
		head[i].sacount,			/* 2 */
		head[i].fincount,			/* 3 */
		head[i].finssent,			/* 4 */
		head[i].rstcount,			/* 5 */
		head[i].tcpfrom,			/* 6 */
		head[i].tcpto,				/* 7 */
		head[i].icmpcount, 			/* 8 */
		head[i].flags, 				/* 9 */
		head[i].emulehits, 			/* 10 */
		ipds,	/* from scan module */		/* 11 */
		tcpds, 	/* from scan module */		/* 12 */
		head[i].pcount);			/* 13 */

	/* note: port tuple comes after this		   14 */
}

static void  __inline
storePort( struct synnode *n, int dport);

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FALSE
#define FALSE   0
#endif

#ifdef TESTMAIN
char *intoa(unsigned int addr);
#endif

/* frequency analysis for edonkey message bytes
*/
static int ede3[256];	/* store message byte after 1st e3 in payload */
static int edc5[256];	/* ditto on c5 */
static int edd4[256];   /* ditton on d4 */

#include "bytecodes.h"

/* called before list can be used 
 */
void init_synlist (struct synlist *l) 
{
	l->entries = 0;
	l->last = NULL;
	bzero (&l->hashlist[0], THSIZE * sizeof(struct synnode *));
}

/*
 * search_srcsyn - hash list lookup
 *
 * If found we increment various counters.
 *
 * if icmpflag is set, we have an icmp packet as the src packet, not a tcp packet.  
 * in that case, we wish to only increment the icmp error counter associated
 * with the destination IP address.	
 *
 * saflag means we should directly increment the sa counter.
 *
 * ircflag is a hint from the L7 irc parser about setting the irc bitflag.
 *
 * basic idea:
 *	1st deal with fin/reset/icmp in terms of lookup by destination address.
 *	icmp does not cause an insert.
 */ 
void
search_srcsyn(struct t_entry *te, struct synlist *l, int saflag, int ircresult, int icmpflag)
{
	struct synnode *n;
	int cflag = 0;
	int doinsert = 1;		/* assume true */
	struct synnode *h, *prev;
	unsigned int mod;
	int i;
	unsigned long ipsrcaddr = te->src.s_addr;
	unsigned long ipdstaddr = te->dst.s_addr;

	/* TBD: think thru what happens with syn attack
	*/
	/* first search on the ip destination address to increment certain reverse counters
	*/
	n = l->hashlist[mod=(ehash((unsigned char *)&ipdstaddr, 4) % THSIZE)];
	for (; n!= NULL ; n=n->next) {
		if (n && ( n->ipsrc == ipdstaddr)) {
			if (te->finflag) {
				n->fincount++;	/* fin returned to this ip source */
			}
			else if (te->resetflag) {
				n->rstcount++;	/* reset returned to this ip source */
			}
			else if (icmpflag) {
				n->icmpcount++;	/* icmp errors sent to this ip src */
			}
			if (!icmpflag)
				n->tcpto++;	/* tcp pkts to this ip src */
			break;
		}
	}

	/* end of icmp portion, tcp packets only from here on.
	 * in addition, we don't have any more screwballs ideas about
	 * ip src vs ip dst.  we are interested in the packet from the ip src
	 * point of view  ONLY.
	*/
	if (icmpflag)
		return;


	/* search the hash/chain list on the ip src.
	*/
	n = l->hashlist[mod=(ehash((unsigned char *)&ipsrcaddr, 4) % THSIZE)];
	for (; n!= NULL ; n=n->next) {
		if (n && ( n->ipsrc == ipsrcaddr)) {
			doinsert = 0;		/* we found it, we don't insert it */
			n->tcpfrom++;		/* it's a tcp packet from the src */
			if (te->synflag) {
				n->syncount++;
				/* if a syn with an ACK coming from an ip_src, then
				 * we count as syn/ack as well
				 */
				if (saflag) {
					n->sacount++;
				}
			}
			else if (te->finflag) {
				n->finssent++;	/* fin sent from this ip src */
			}
			else if (te->resetflag) {
				; 
			}
			/* l7 connection checks on data packets only
			*/
			else {
				/* irc check -- based on simple lex/scanning -- see ircscan.c
				*/
				if (ircresult) {
					if (ircresult == 1) {
						n->flags |= (IRCJOINS);
					}
					else if (ircresult == 2) {
						n->flags |= (IRCPINGS);
					}
					else if (ircresult == 3) {
						n->flags |= (IRCPONGS);
					}
					else if (ircresult == 4) {
						n->flags |= (IRCPRIVMSG);
					}
				}
				/* bittorrent/gnutella etc... using bcmp at start of L7 payload
				 * if you have done this once per sample period do NOT
				 * do it again.
				*/
				if (te->psize >= 20) {
					/* bittorrent flow?  etc.
					*/
					if (((n->flags & BITTORRENT) == 0) && (bcmp(te->p, btcode, 20) == 0)) {
						n->flags |= (BITTORRENT);
					}
					else if (((n->flags & GNUTELLA) == 0) && (bcmp(te->p, gnucode, 17) == 0)) {
						n->flags |= (GNUTELLA);
					}
					else if (((n->flags & MORPHEUS) == 0) && (bcmp(te->p, morphcode, 20) == 0)) {
						n->flags |= (MORPHEUS);
					}
					else if (((n->flags & KAZAA) == 0) && (bcmp(te->p, kazaacode, 20) == 0)) {
						n->flags |= (KAZAA);
					}
				}
				/* edonkey check.  binary statistical guess
				*/
				if (te->psize >= 6) {
					unsigned char c = *te->p;
					unsigned char c2 = *(te->p+5);	/* message type */

					/* 2nd byte chosen on basis of brief stat. study
					*/
					/* if edonkey payload byte */
					if (c == 0xe3) {
						/* 0x01 is hello, client-server
						 * 0x46 is msg sending part, client-client
						 */
						if (c2 == 0x01) { 
							n->emulehits++;
							n->flags |= (EDONKEY);
						}
#ifdef NOTDEF
						else if (c2 == 0x46) {
							/*
							n->flags |= (EDONKEY1);
							*/
							n->emulehits[1]++;
						}
#endif
					}
#ifdef NOTDEF
					/* if emule payload byte */
					else if (c == 0xc5) {
						/* 0x02 - hello answer
						 * 0x60 - queue ranking
						 */
						if (c2 == 0x02) {
							n->emulehits[2]++;
						}
						else if (c2 == 0x60) {
							n->emulehits[3]++;
						}
					}
#endif
#ifdef EMULESTATS
					if (c == 0xe3) {
						ede3[c2]++;
					}
					else if (c == 0xc5) {
						edc5[c2]++;
					}
					else if (c == 0xd4) {
						edd4[c2]++;
					}
#endif
				}
			}  /* end data packet portion */
			/* a few more flag checks
			*/
			if (te->sport == 80 || te->sport == 443) {
				n->flags |= (WEBSERVER);
			}	
			else if (te->dport == 25) {
				n->flags |= (EMAIL);
			}
			/* do dst port sampling
			*/
			storePort(n, te->dport);
			break;
		}
	}

	/* we found it therefore we do not insert it */
	if (doinsert == 0) {
		return;
	}

	/* insert needed */

	n = (struct synnode *) malloc(sizeof(struct synnode));
	if (n == NULL) {
		syslog(LOG_ERR, "malloc failure in insert_srcsyn, entry/count %d",
			l->entries);
		return;
	}
	bzero(n, sizeof (struct synnode));
	l->entries += 1;
	n->tcpfrom = 1;
	n->ipsrc = ipsrcaddr;
	
	/* mark ports unused */
	for ( i = 0; i < MAXPORTS; i++) {
		n->dports[i] = -1;
	}
	if (te->synflag) {
		n->syncount = 1;
		if (saflag) {
			n->sacount = 1;
		}
		storePort(n, te->dport);	
	}
	else if (te->finflag) {
		n->finssent = 1;
		n->tcpto = 1;
		storePort(n, te->dport);	
	}
	else if (te->resetflag) {
		/* not doing anything with resets */
	}
	else {
		storePort(n, te->dport);	
		if (ircresult) {
			if (ircresult == 1) {
				n->flags |= (1 << IRCJOINS);
			}
			else if (ircresult == 2) {
				n->flags |= (1 << IRCPINGS);
			}
			else if (ircresult == 3) {
				n->flags |= (1 << IRCPONGS);
			}
		}
	}

	/* insert the synnode in the hash list
	*/
	n->next = NULL;
	h = l->hashlist[mod=(ehash((unsigned char *)&ipsrcaddr, 4) % THSIZE)];
	if (h == NULL) {
		l->hashlist[mod] = n;
	}
	else {
		n->next = h;
		l->hashlist[mod] = n;
	}
}

/*
 * extract work weight associated with ip src address
 */

int
searchSyn(struct synlist *l, unsigned long ipsrcaddr)
{
	struct synnode *n;
	unsigned int mod;
	int t1;

	/* first search on the ip destination address to increment certain reverse counters
	*/
	n = l->hashlist[mod=(ehash((unsigned char *)&ipsrcaddr, 4) % THSIZE)];
	for (; n!= NULL ; n=n->next) {
		if (n && ( n->ipsrc == ipsrcaddr)) {
			t1 = n->tcpfrom + n->tcpto;
			if (t1 == 0)
				return(0);
			/* return the work weight as an integer between 0..100 */
			t1 = ((n->syncount + n->finssent + n->rstcount) / (t1 * 1.0)) * 100;                        
			return(t1);
		}
	}
	return(0);
}

static void __inline
storePort(struct synnode *n, int dport)
{
	int i;

	for ( i = 0; i < MAXPORTS; i++) {
		/* have not seen the port and we have room,
		 * therefore store it and increment counter and return
		 */
		if (n->dports[i] == -1) {
			n->dports[i] = dport;
			n->dpcounts[i] = 1;
			n->pcount++;
			return;
		}
		/* found it, increment count, and return
		*/
		if (n->dports[i] == dport) {
			n->dpcounts[i]++;
			return;
		}
	}
	/* didn't find it, must not be room, 
	 * too bad 
	 */
	return;
}

/*
 * deallocate malloced hash entries
 */

void
destroy_synlist(struct synlist *l) 
{
	int i;
	struct synnode *n;
	struct synnode *prev;
	for (i =0; i< THSIZE; i++) {
		n = l->hashlist[i];
		while(n != NULL) {
			prev = n;
			n = n->next;
			free(prev);
		}
	}
}

#ifdef TESTMAIN
static
dump_hash(struct synlist *l) {
	struct synnode *h;
	int counters[THSIZE];
	int i;
	int ls = 0;
	printf("No. of entries %d\n", l->entries);
	bzero(&counters, sizeof(int)*THSIZE);
	printf("\nhashdump---------\nlist entries %d\n",l->entries);
	for (i=0;i<THSIZE;i++) {
		h = l->hashlist[i];
		while(h != NULL) {
			ls = 1;
			printf("%d:", i);
			printf("ip src %x", intoa(h->ipsrc));
			printf(":s/f/r: %d %d %d", h->syncount,
				h->fincount, h->rstcount);
			counters[i]++;
			h = h->next;
		}
		if (ls) {
			printf(":%d \n",i);
		}
		ls = 0;
	}
	printf("\n");
	for (i = 0; i<THSIZE; i++) {
		if (counters[i] != 0) {
			printf("counters[%d]:%d buckets\n", i, counters[i]);
		}
	}
	printf("end hash dump---------\n");
}
#endif

/* pack the hashlist into an array
 * and return the array.
 *
 *	returns:
 *		1 - packed array exists
 *		0 - malloc failure
 */
static
struct synnode *
pack_array(struct synlist *l)
{
	struct synnode *head, *n;
	unsigned int c = l->entries;  /* count of hash entries allocated */
	int i = 0;
	int z;
	int hcounter = 0;
	WEIGHTCOUNTER t1;
	head = (struct synnode *) malloc(c * sizeof (struct synnode));
	if (head == NULL) {
		syslog(LOG_ERR, "malloc failure in syn/pack_array, count %d",
			l->entries);
		return (NULL);
	}
	/* walk the hashlist
	*/
	for (i =0 ; i<THSIZE; i++) {
		n = l->hashlist[i];
		/* if n is non-nulln use, therefore put it in the array,
		* and increment hcounter
		*/
		while (n != NULL) {
			head[hcounter].ipsrc    = n->ipsrc;
			head[hcounter].syncount = n->syncount;
			head[hcounter].sacount  = n->sacount;
			head[hcounter].fincount = n->fincount;
			head[hcounter].finssent = n->finssent;
			head[hcounter].rstcount = n->rstcount;
			head[hcounter].tcpfrom  = n->tcpfrom;
			head[hcounter].tcpto    = n->tcpto;
			head[hcounter].icmpcount = n->icmpcount;
			head[hcounter].pcount   = n->pcount;
			head[hcounter].flags    = n->flags;
			head[hcounter].emulehits = n->emulehits;

			t1 = n->tcpfrom + n->tcpto;
			/* compute the work weight
			*/
			if (t1 == 0) {
				head[hcounter].ww = head[hcounter].work = 0;
			}
			else {
				head[hcounter].ww = head[hcounter].work = 
					(n->syncount + n->finssent + n->rstcount) / t1;                        
				/* sort with w=100% (w varies from 0..1 here) by syn count by simply adding the syncount back in.
				 * it would be possible to do this within all weights by scaling
				 * the syncount by an absolute max. syns per second but it isn't
				 * clear as to whether or not that is needed.
				*/
				if (head[hcounter].work >= 1) {
					head[hcounter].work += n->syncount;
				}
			}

			for (z = 0; z < n->pcount; z++) {
				head[hcounter].dports[z] = n->dports[z]; 
				head[hcounter].dpcounts[z] = n->dpcounts[z];
			}
			n = n->next;
			hcounter++;
		}
	}  
	return(head);
}

static
int
sort_by_syncount(e1,e2)
struct synnode *e1, *e2;
{
        if (e1->syncount > e2->syncount) return -1;
        if (e1->syncount < e2->syncount) return 1;
        return 0;
}

static
int
sort_by_work(e1,e2)
struct synnode *e1, *e2;
{
        if (e1->work > e2->work) return -1;
        if (e1->work < e2->work) return 1;
        return 0;
}

/*
static
sort_by_scalar(e1,e2)
int *e1, *e2;
{
        if (*e1 > *e2) return -1;
        if (*e1 < *e2) return 1;
        return 0;
}
*/

/* given a synlist, sort it, and print it out (count entries max)
 *
 * inputs:
 *      fd - file for print output
 *      l - hash list
 *      count - max desired outputs for all filters
 */
void
print_synlist(FILE *fd, struct synlist *l, int count, int wflag, 
char *wormfile, int ipaflag, unsigned long usnet, unsigned long usmask, int sflag,
struct slist *isl, struct slist *tsl)
{
	extern struct topn_syn topn_syn;
	struct synnode *head;
	int i;
	int z;
	int c = l->entries;  /* no of hash elements */
	int us, them, wcount, wormcount, p2pcount;
	FILE *wfd = NULL;
	FILE *p2pfd = NULL;
	FILE *sfd = NULL;
	int tcpworm;
	COUNTER ipdst_hits = 0;	
	COUNTER tcpdst_hits = 0;
	WEIGHTCOUNTER synww = 0;
	WEIGHTCOUNTER wormww = 0;
	WEIGHTCOUNTER p2pww = 0;
	
	/* open tcpworm file for writing
	*/
	if (wflag) {
		wfd = fopen(wormfile, "w");
		if (wfd == NULL) {
			syslog(LOG_ERR, "could not open %s:(%m)", wormfile);  
			exit(1);
		}
	}
	us = them = wcount = p2pcount = wormcount = 0;

	/* open p2p file for writing
	*/
	if (topn_syn.p2pflag) {
		p2pfd = fopen(topn_syn.p2pfile, "w");
		if (p2pfd == NULL) {
			syslog(LOG_ERR, "could not open %s:(%m)", topn_syn.p2pfile);  
			exit(1);
		}
	}

	/* form the hashlist into a packed array
	*/
	head = pack_array(l);
	if (head == NULL) {
		return;
	}
	/* quicksort it either by syn count or by adjusted worms at the top syncount
	 * sortbysyn is the default
	*/

	if (topn_syn.sortbysyn) {
		qsort((void *)head, c, sizeof(struct synnode), sort_by_syncount);
	}
	else {
		qsort((void *)head, c, sizeof(struct synnode), sort_by_work);
	}

	/* loop thru the syn list and get the top N synners.
	 * the new raw format is as follows:
	 * ip: syn: sa: fin: finssent: rst: tcpfrom: tcpto: icmpcount: L7 flags: ehits: 
	 * 	Nports: port,count,port,count,port,count
	*/
	fprintf(fd, "syn_list: %u : ", l->entries); 
	for (i = 0; i < l->entries; i++) {

		/* produce top N syn tuple, stop at count
		 * this goes into mon.lite file.
		*/
		if ( i < count) {
			/* print to mon.lite
			*/
			printSynTuple(fd, head, i, sflag, isl, tsl);

			for (z = 0; z < head[i].pcount; z++) {
				fprintf(fd, "%u,%u,", head[i].dports[z], head[i].dpcounts[z]);
			}
			fprintf(fd, ":");
		}
	
		synww += head[i].ww;

		/*
		 * from looking at the syndump.txt file (all syn list entries sorted)
		 * it appears that 20-40 is a reasonable range.  
		 * WORMWEIGHT
		 */

		/* syncount is unsigned and might be zero and we also want to catch cases of
		 * all fins but no syns, as well as no syns/fins, but we have resets.
		 */
		tcpworm = 0; 
		if (head[i].syncount <= head[i].fincount) {
			;
		}
		else if (((head[i].syncount - head[i].fincount) > topn_syn.wormdiff) ) {
			tcpworm = 1;
		}
		/* mucho FINS sent, but not so much syns sent: nmap -sF */
		if (!tcpworm && (head[i].syncount <= head[i].finssent)) {
			if ((head[i].finssent - head[i].syncount) > topn_syn.wormdiff)  
				tcpworm = 1;
		}
		/* more than wormdiff packets sent and there exist resets.  this catches
		 * nmap scans with acks or just pushes(data) but no control
		 * TBD: possibly better if: rstcount > wormdiff and that's it?
		 * this version is looser.
		 */
		if ((head[i].syncount == 0) && (head[i].rstcount > topn_syn.wormdiff)
			&& (head[i].tcpfrom > topn_syn.wormdiff)) {
				tcpworm = 1;
		}
	
		if (tcpworm) {
			/* if the true work weight is greater than the config pre-filter
			 * weight, then we count things in tworm.  
			 * default: topn_syn.weight is 0.
			 */
			if ((int)(head[i].ww * 100) >= topn_syn.weight) { 
				wcount++;	/* count of worms acc. to topn_syn.weight */
				if (ipaflag) {
					if (((head[i].ipsrc & usmask) == usnet)) {
						us++;
					}
					else  { 
						them++;
					}
				}
			}
			/* print to wormfile
			*/
			printSynTuple(wfd, head, i, sflag, isl, tsl);
			for (z = 0; z < head[i].pcount; z++) {
				fprintf(wfd, "%u,%u,", head[i].dports[z], head[i].dpcounts[z]);
			}
			wormww += head[i].ww;
			wormcount++;	/* total count of subset hosts in tcpworm.txt */

			fprintf(wfd, ":\n");
		}
		/* if p2p file, write all p2p tuples into it.
		*/
		if (p2pfd) {
			if (head[i].flags & P2PBITS) {
				printSynTuple(p2pfd, head, i, sflag, isl, tsl);
				for (z = 0; z < head[i].pcount; z++) {
					fprintf(p2pfd, "%u,%u,", head[i].dports[z], head[i].dpcounts[z]);
				}
				fprintf(p2pfd, ":\n");
				p2pcount++;
				p2pww += head[i].ww;
			}
		}
	}


#ifdef SYNDUMP
	/* this is used in batch mode to study what has showed up in the syn list
	*/
	sfd = fopen("/tmp/syndump.txt", "w");
	if (sfd == NULL) {
		syslog(LOG_ERR, "could not open %s:(%m)", "/tmp/syndump.txt");  
		exit(1);
	}
	/* write all the syn info to a file -- all entries
	*/
	for (i = 0; i < l->entries; i++) {
			printSynTuple(sfd, head, i, sflag, isl, tsl);
			for (z = 0; z < head[i].pcount; z++) {
				fprintf(sfd, "%u,%u,", head[i].dports[z], head[i].dpcounts[z]);
			}
			fprintf(sfd, ":\n");
	}
	fclose(sfd);
#endif
#ifdef EMULESTATS
	/* this is just for gathering of stats for edonkey and has no runtime use
	*/
	{
		FILE *sfd;
		sfd = fopen("/tmp/emulestats.txt", "w");
		if (sfd) {
			fprintf(sfd, "C5C5c5 codes\n");
			for (i = 0; i < 256; i++) {
				fprintf(sfd, "(%x: %x) ", i, edc5[i] & 0xff);
				if (i % 16 == 0)
					fprintf(sfd, "\n");
			}
			fprintf(sfd, "\nE3E3e3 codes\n");
			for (i = 0; i < 256; i++) {
				fprintf(sfd, "(%x: %x) ", i, ede3[i] & 0xff);
				if (i % 16 == 0)
					fprintf(sfd, "\n");
			}
			fprintf(sfd, "\nD4D4d4 codes\n");
			for (i = 0; i < 256; i++) {
				fprintf(sfd, "(%x: %x) ", i, edd4[i] & 0xff);
				if (i % 16 == 0)
					fprintf(sfd, "\n");
			}
			fclose(sfd);
		}
	}
#endif
	
	fprintf(fd, "\n");
	if (ipaflag) {
		fprintf(fd, "tworm: %d: %d: %d:\n", wcount, us, them);
	}

	/* if we have a tcpworm file also do netww RRDTOOL graph stats
	 * these are average work weights for various classes of behavior
	 */
	if (wfd) {
#ifdef NETWW
                /* rethink this, probably the syn tuple should not include
                 * the other tuple sets.
                */
		WEIGHTCOUNTER syn_avg, worm_avg, p2p_avg;

		/* produce netww tuple, avoid divide by zero
		 */
		if (c != 0) {
			syn_avg = synww / c;
		}
		else {
			syn_avg = 0;
		}
		if (wormcount != 0) {
			worm_avg = wormww / wormcount;
		}
		else {
			worm_avg = 0;
		}
		if (p2pcount != 0) {
			p2p_avg = p2pww / p2pcount;
		}
		else {
			p2p_avg = 0;
		}

		/* print average work weights for syn tuple, worm, and p2p.
 		 * format:  syntuple count:  syntuple avg ww, worm count, worm avg work weight,
		 *	p2pcount,  p2p avg work weight
		 */
		fprintf(fd, "netww: %u: %u: %u: %u: %u: %u:\n",
			c, 
			(int)((syn_avg) * 100), 
			wormcount, 
			(int)((worm_avg) * 100), 
			p2pcount, 
			(int)((p2p_avg) * 100));
#endif
		fclose(wfd);
	}
	if (p2pfd) {
		fclose(p2pfd);
	}

	free(head);

        /* check for trigger condition
	 * we trigger if the total count is greater than the threshold.
        */
	trigger_condition_check(TWORM, fd, wcount, "tworm");
}          


#ifdef TESTMAIN

this code is old and broken

static int f = 0;

unsigned long
getRandomWord()
{
        static unsigned long rw;

#define REALLYRANDOM
#ifdef REALLYRANDOM
        if (f == 0) {
                f = open("/dev/urandom", 0);
                if (f < 0) {
                        perror("open");
                        exit(1);
                }
        }
        read(f, &rw, sizeof(unsigned long));
#else
        rw = rand();
#endif
        return(rw);
}


functionTest()
{
	struct synlist l;
	unsigned long ipsrc;
	int rc;
	struct synnode *n = NULL;

	/* load sample ipkey and counter
	*/
	ipsrc = inet_addr("1.1.1.1");

	init_synlist(&l);

	n = search_srcsyn(NULL, &l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("search failed\n");
	} 
	
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	n = search_srcsyn(NULL, &l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 

	/* add 4 more nodes and dump the list
	*/
	ipsrc = inet_addr("1.1.1.2");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	ipsrc = inet_addr("1.1.1.3");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}
	ipsrc = inet_addr("1.1.1.4");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	ipsrc = inet_addr("1.1.1.5");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	ipsrc = inet_addr("1.1.1.6");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	/* now search for a given node, and increment its
	 * value
	 */
	n = search_srcsyn(&l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 
	else {
		printf("searchNode found node\n");
	}

	/* now search for a previous node, note last
	 * as a simple cache should not be used here
	 */
	ipsrc = inet_addr("1.1.1.4");
	n = search_srcsyn(&l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 
	else {
		printf("searchNode found node\n");
	}

	dump_hash(&l);

	print_synlist(stdout, &l, 10);

	destroy_synlist(&l);
}


/*
 * borrowed from trafshow.
 *
 * A faster replacement for inet_ntoa().
 */

char *
intoa(addr)
unsigned int addr;
{
	register char *cp;
	register unsigned int byte;
	register int n;
	static char buf[sizeof(".xxx.xxx.xxx.xxx")];

	addr = ntohl(addr);
	cp = &buf[sizeof buf];
	*--cp = '\0';

	n = 4;
	do {
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if (byte > 0) {
			*--cp = byte % 10 + '0';
			byte /= 10;
			if (byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);

	return cp + 1;
}


main()
{
	functionTest();
}
#endif


